* PDFClass.PRG
* (c) all code Lisa Slater Nicholls >L<
* revision of PDFListener for earlier versions of VFP, tested down to VFP 5

* --------------- testcode starts here

#DEFINE PICKREPORT_LOC "Add a report -- " + CHR(13) + ;
                       IIF(TYPE([SET("REPORTB")]) = "N", ;
                           "must not have a Printer Environment saved", ;
                           "doesn't matter whether it has Printer Environment or not ") + ;
                       CHR(13) + ;
                       "(or ESC to finish)..."
#DEFINE PDFAVAILABLE_LOC thisPDF.TargetFileName + " is now available."
#DEFINE ERRORPREFIX_LOC "ERROR: "

#DEFINE BASE_GS_LOCATION  "GS\"

#DEFINE TESTPRINTSETUP_ONLY .F.

EXTERNAL TABLE VFP5Install.DBF

LOCAL loX, lcDir, liSetReport

#IF TYPE([SET("REPORTB")]) = "N" 
   IF SET("REPORTB") > 89
      liSetReport = SET("REPORTB")
      SET REPORTBEHAVIOR 80
   ENDIF      
#ENDIF   

lcDir = SET("DIRECTORY")
CD ( SUBSTR(SYS(16,1),1,RAT("\",SYS(16,1))))
loX = CREATEOBJECT("test")
? loX.Run()
CD (lcDir)
#IF TYPE([SET("REPORTB")]) = "N" 
   IF NOT EMPTY(liSetReport)
      SET REPORTBEHAVIOR liSetReport
   ENDIF
#ENDIF   
RETURN


DEFINE CLASS Test AS Custom

    outputFile = SYS(2015)
    oldStyle = .T.
    showPDF = .T.
    removeReportsAfterRun = .T.
    
    PROCEDURE TestPrintSetupCreation(toPDF)
      RETURN toPDF.VerifyPrinterSetup()
    ENDPROC
    
   
    PROCEDURE Run
       LOCAL thisPDF, cReport, lRun, cpass
       thisPDF = CREATEOBJECT("pdfClass")
       IF TESTPRINTSETUP_ONLY
         RETURN THIS.TestPrintSetupCreation(thisPDF)
       ENDIF         
       * where you want ghostscript files to be 
       * installed (all tempfiles will go here too)
       * thisPDF.GSLOCATION = FULLPATH(".\" + BASE_GS_LOCATION)
       * wherever you want the result
       * thisPDF.targetFileName = SET("DIRECTORY")+ "\" + THIS.outputFile
       * what you want the driver name to look like in Windows
        thisPDF.PSDRIVERSETUPNAME = "My New Application PDF" 

       * change other attributes, such as:
       thisPDF.RETAINTEMPFILES = .T.  && will go into the recycle bin
                                
       * thispdf.QuietMode = .T.    && you don't even want to see status msgs

      WAIT WINDOW NOWAIT PICKREPORT_LOC
      
      cReport = GETFILE("FRX")
      
      DO WHILE NOT EMPTY(cReport)
         lRun = .T.
         thisPDF.addReport(cReport)
         * NB: you can add explicit clauses:
         * thisPDF.addReport(cReport"," NEXT 1") && etc
         WAIT WINDOW NOWAIT PICKREPORT_LOC
         cReport = GETFILE("FRX")
      ENDDO
      
      WAIT CLEAR
      
      IF lRun THEN

         *&* thisPDF.runReports(this.removeReportsAfterRun, this.oldStyle)
         * VFP 9 runReports arguments: tlRemoveReportsAfterRun, tlOmitListenerReferences
         thisPDF.runReports(this.removeReportsAfterRun)         
         IF LEN(thisPDF.GetLastErrorMessage()) > 0 THEN
            MESSAGEBOX(ERRORPREFIX_LOC+ thisPDF.GetLastErrorMessage())
         ELSE
            IF this.ShowPDF
               DECLARE INTEGER ShellExecute ;
                   IN SHELL32.DLL ;
                   INTEGER nWinHandle,;
                   STRING cOperation,;   
                   STRING cFileName,;
                   STRING cParameters,;
                   STRING cDirectory,;
                   INTEGER nShowWindow
               ShellExecute(0,"Open",thisPDF.targetFilename,"","",1)
               CLEAR DLLS ShellExecute
            ELSE
               MESSAGEBOX(PDFAVAILABLE_LOC)  
            ENDIF   
         ENDIF
         
      ENDIF   
      RETURN lRun
      
   ENDPROC

ENDDEFINE   

* ---------------------- PDFClass code proper starts here


#DEFINE WIN_API_ERROR_LOC "Error calling Windows API: "
#DEFINE INSTALLING_DRIVER_LOC "Installing PDF Driver... "
#DEFINE INSTALLING_LIBRARY_LOC "Installing PDF Libraries... "
#DEFINE GENERAL_FAILURE_LOC "Could not create PDF file."
#DEFINE PROCESS_FAILURE_LOC "Could not create GS process."
#DEFINE OUTPUTPDF_APPNAME_LOC "PDF Class"
#DEFINE PSETUP_DEFAULTNAME_LOC "VFP-PDF"


*&* from VFP 9 superclasses:
#DEFINE OUTPUTCLASS_ERRNOLABEL_LOC           "Error:           "
#DEFINE OUTPUTCLASS_ERRPROCLABEL_LOC         "Method:       "
#DEFINE OUTPUTCLASS_ERRLINELABEL_LOC         "Line:            " 
#DEFINE OUTPUTCLASS_FILENAME_CHARS_DISALLOWED  [?*"<>|]

*&* PDFListener/PDFClass-specific:
#DEFINE VFP_USER_DEFINED_ERROR  1098
#DEFINE MB_ICONEXCLAMATION      48     
#DEFINE MB_ICONSTOP             24     

#DEFINE DRIVER_TO_USE "Apple Color LaserWriter 12/600"
#DEFINE BASE_GS_EXEC_FILE "gswin32c.exe"

#DEFINE GS_COMMAND_STRING_1  " -q -dNOPAUSE -I./lib;./fonts " + ;
                             "-sFONTPATH=./fonts " + ;
                             "-sFONTMAP=./lib/FONTMAP.GS " + ;
                             "-sDEVICE=pdfwrite -sOUTPUTFILE="
#DEFINE GS_COMMAND_STRING_2 " -dBATCH  "   

DEFINE CLASS PDFClass as Custom
                                       
   PROTECTED FRXPrintInfoAlias, ReportTempFiles[1], ;
             ReportFileNames[1], ReportClauses[1]

   TargetFileExt = "PDF"
   HadError = .F.
   LastErrorMessage = ""
   QuietMode = .F.
   AppName =  OUTPUTPDF_APPNAME_LOC
   TargetFileName = SYS(2023) + "\P"+SYS(2015)
   

   GSLocation = SYS(2023) + "\" + BASE_GS_LOCATION 
   GSExecFile = BASE_GS_EXEC_FILE
   PSDriverSetupName = PSETUP_DEFAULTNAME_LOC
   FRXPrintInfoAlias = "P"+SYS(2015)
   retainTempFiles = .F.   
   WaitAttempts = 20
   WaitInterval = 200
   
   
   PROCEDURE Error(nError, cMethod, nLine)
     THIS.DoError(nError, cMethod, nLine)
   ENDPROC
   
   PROTECTED PROCEDURE DoError(nError, cMethod, nLine)
      *&* adapted from VFP 9 superclass Error behavior
      LOCAL m.lcOnError,m.lcErrorMsg,m.lcCodeLineMsg
      THIS.HadError = .T.
      m.lcOnError=UPPER(ALLTRIM(ON("ERROR")))
      IF NOT EMPTY(m.lcOnError)
         m.lcOnError=STRTRAN(STRTRAN(STRTRAN(lcOnError,"ERROR()","nError"), ;
			"PROGRAM()","cMethod"),"LINENO()","nLine")
         &lcOnError
         RETURN
      ENDIF
      m.lcErrorMsg = THIS.PrepareErrorMessage(m.nError,m.cMethod, m.nLine)
      THIS.LastErrorMessage = m.lcErrorMsg

      MESSAGEBOX(m.lcErrorMsg, MB_ICONSTOP )
   ENDPROC
   
   PROCEDURE GetLastErrorMessage() 
     RETURN THIS.LastErrorMessage
   ENDPROC
   
  
*!*   VFP 9 PROCEDURE RunReports(tlRemoveReportsAfterRun, tlOmitListenerReferences)
  PROCEDURE RunReports(tlRemoveReportsAfterRun)
     IF (NOT ISNULL(THIS.ReportFileNames)) AND ;
         (NOT EMPTY(THIS.ReportFileNames[1])) AND ;
         THIS.LoadPrinterInfo()
         THIS.TargetFileName = ALLTR(THIS.TargetFileName)
         DO CASE
         CASE RAT(".",THIS.TargetFileName) = 0 
            THIS.TargetFileName = THIS.TargetFileName + "." + THIS.TargetFileExt
         CASE RAT(".",THIS.TargetFileName) = LEN(THIS.TargetFileName)         
            THIS.TargetFileName = THIS.TargetFileName + THIS.TargetFileExt            
         ENDCASE
         THIS.verifyTargetFile()
         THIS.DoRunReports()
         THIS.UnloadPrinterInfo()
         RETURN THIS.ProcessPDF(tlRemoveReportsAfterRun)
      ELSE
         RETURN .F.   
      ENDIF   
   ENDPROC
   
   PROTECTED PROCEDURE DoRunReports()
      *&* adapted from VFP 9 superclass RunReports behavior
      IF NOT EMPTY(THIS.ReportFileNames[1])
         LOCAL m.liIndex, m.lcClauses
         FOR m.liIndex = 1 TO ALEN(THIS.ReportFileNames)
             m.lcClauses = UPPER(THIS.ReportClauses[m.liIndex])
             REPORT FORM (THIS.ReportFileNames[m.liIndex]) &lcClauses  
          ENDFOR
      ENDIF   
   ENDPROC
        
   PROTECTED PROCEDURE ProcessPDF(tlRemoveReportsAfterRun)
      IF EMPTY(THIS.reportTempFiles[1])
         RETURN .F.
      ELSE
         * move the collection of reportTempFiles into a GS batch file
         * for processing
         LOCAL llReturn, lcCmdFile, lcCmd, lcDir
         lcCmdFile = THIS.MakeGSCommandFile()
         * shell out to gs      
         lcCmd = THIS.GSLocation + THIS.GSExecFile+ " @"+ SUBSTR(lcCmdFile,RAT("\",lcCmdFile) + 1)
         lcCmd = STRTRAN(lcCmd,"\","/")
         lcDir = SET("DIRECT")
         CD (THIS.GSLocation)
         llReturn = THIS.ProgExecute(lcCmd," NOT FILE(["+THIS.TargetFileName+"]) ",THIS.WaitAttempts)      

         CD (lcDir)
         IF EMPTY(THIS.LastErrorMessage) AND NOT FILE(THIS.TargetFileName)
            THIS.LastErrorMessage = ;
              THIS.PrepareErrorMessage( ;
                  VFP_USER_DEFINED_ERROR, ;
                 PROGRAM(), ;
                 0, ;
                 THIS.AppName, ;
                 GENERAL_FAILURE_LOC+" "+THIS.TargetFileName,;
                 "")
         ENDIF        
         THIS.CleanupTempFiles(lcCmdFile)       
         IF tlRemoveReportsAfterRun
            THIS.RemoveReports()
         ENDIF
         RETURN llReturn
      ENDIF   
   ENDPROC    
      
   PROTECTED PROCEDURE CleanupTempFiles(tcFile)
      LOCAL liIndex, lvFile, lcRetain
      IF THIS.retainTempFiles
         lcRetain = " RECYCLE"
      ELSE
         lcRetain = ""
      ENDIF
      FOR liIndex = 1 TO ALEN(THIS.reportTempFiles)
          lvFile = THIS.reportTempFiles[liIndex]
          IF (NOT EMPTY(lvFile)) AND FILE(lvFile)
             ERASE (lvFile) &lcRetain
          ENDIF   
      ENDFOR
      IF (NOT EMPTY(tcFile)) AND FILE(tcFile)
         ERASE (tcFile) &lcRetain
      ENDIF
   ENDPROC
   
*!*   VFP 9 PROCEDURE AddReport(tcFRXName, tcClauses, toListener)
  PROCEDURE AddReport(tcFRXName, tcClauses)
     LOCAL liReports
     IF ISNULL(THIS.reportClauses) OR EMPTY(THIS.reportClauses[1])
        liReports = 0
     ELSE
        liReports = ALEN(THIS.reportClauses)
     ENDIF
*!*	     DODEFAULT(tcFRXName,tcClauses,toListener)
     THIS.DoAddReport(tcFRXName, tcClauses)
     IF (NOT ISNULL(THIS.reportClauses)) AND ;
        ALEN(THIS.reportClauses) > liReports
        THIS.AdjustReportClauses(liReports + 1)
     ENDIF
   ENDPROC
   
   PROTECTED PROCEDURE DoAddReport(tcFRXName, tcClauses)
      *&* adapted from VFP 9 superclass AddReport behavior --
      *&* we don't have collections here, just arrays,
      *&* and we don't have a listener reference to match each report
      IF TYPE("m.tcFrxName") = "C" AND ;
        (FILE(m.tcFRXName) OR FILE(m.tcFRXName + ".FRX") OR FILE(m.tcFRXName + ".LBX"))

         * If any is null or , create all 
         * always add to all 
         * to keep them in synch
         
         LOCAL liCount
   
         IF EMPTY(THIS.ReportFileNames[1]) OR ;
            EMPTY(THIS.ReportClauses[1]) 
            * start fresh
            liCount = 1
         ELSE
            liCount = ALEN(THIS.ReportFileNames) + 1
         ENDIF
         
         DIME THIS.ReportFileNames[liCount]
         DIME THIS.ReportClauses[liCount]

         THIS.ReportFileNames[liCount] = m.tcFRXName
         IF TYPE("m.tcClauses") = "C"
            THIS.ReportClauses[liCount] = m.tcClauses
         ELSE
            THIS.ReportClauses[liCount] = ""
         ENDIF

     ELSE

        * TBD: should we error here?   
   
     ENDIF   

   ENDPROC
   
   PROTECTED PROCEDURE RemoveReports()
*!*	      DODEFAULT()
      THIS.DoRemoveReports()
      DIMENSION THIS.reportTempFiles[1]
      THIS.reportTempFiles[1] = .F.
   ENDPROC
   
   PROTECTED PROCEDURE DoRemoveReports()
      *&* adapted from VFP 9 superclass RemoveReports behavior
    DIME THIS.ReportFileNames[1]
    DIME THIS.ReportClauses[1]
    STORE .F. TO THIS.ReportFileNames, THIS.ReportClauses
   ENDPROC
   
   PROTECTED PROCEDURE AdjustReportClauses(tiWhich)
      LOCAL lcClauses, lcTempFile, laWords[1], liIndex, lcWord
      IF EMPTY(THIS.reportClauses[tiWhich])
         lcClauses = ""
      ELSE
         lcClauses = NVL(THIS.reportClauses[tiWhich],"")
      ENDIF
      lcTempFile = THIS.GSLocation + "PS"+SYS(2015) + ".PS"
      * parse out the possible PREVIEW/TO PRINT
      lcClauses = " " + UPPER(lcClauses)
      IF " TO " $ lcClauses
         * TO PRINT or TO FILE <name>, where FILE is optional
         lcClauses = STRTRAN(lcClauses,"FILE","")
      ENDIF
      *&*&*& VFP 9 behavior handling error-assessment of 
      *&* various clauses has been removed here.
      *&* you should add your own parsing code to get rid of
      *&* TO, PREVIEW, PROMPT if you aren't sure what clauses
      *&* you might have been passed

      * Now adjust the clauses to go TO FILE
      * with a temporary name
      lcClauses = " "+lcClauses + " "
      
      *&* NODIALOG is not available in every possible
      *&* version -- you can add the following back in
      *&* if you can use it, but I'm not version-testing here
      
*!*	      IF ATC(" NODI ",lcClauses) = 0 AND ;
*!*	         ATC(" NODIA ",lcClauses) = 0 AND ;
*!*	         ATC(" NODIAL ",lcClauses) = 0 AND ;
*!*	         ATC(" NODIALO ",lcClauses) = 0 AND ;
*!*	         ATC(" NODIALOG ",lcClauses) = 0 
*!*	         lcClauses = lcClauses + " NODIALOG "
*!*	      ENDIF

      IF ATC(" NOCO ",lcClauses) = 0 AND ;
         ATC(" NOCON ",lcClauses) = 0 AND ;
         ATC(" NOCONS ",lcClauses) = 0 AND ;
         ATC(" NOCONSO ",lcClauses) = 0 AND ;
         ATC(" NOCONSOL ",lcClauses) = 0 AND ;
         ATC(" NOCONSOLE ",lcClauses) = 0          
         lcClauses = lcClauses + " NOCONSOLE "
      ENDIF
         
      THIS.reportClauses[tiWhich] = ;
         lcClauses + " TO FILE (["+ UPPER(lcTempFile) + "]) " 

      * add to the collection of temporary Filenames
      IF tiWhich > 1
         DIMENSION THIS.reportTempFiles[tiWhich]
      ENDIF   
      THIS.reportTempFiles[tiWhich] = lcTempFile
   ENDPROC
   
   
   PROTECTED PROCEDURE MakeGSCommandFile()
      LOCAL lcFile, lcContents, lcFileString, liH
      lcFile = THIS.GSLocation + "C"+SYS(2015)+".TXT"
      lcFileString = THIS.GetQuotedFileString()
      lcContents = GS_COMMAND_STRING_1 + ;
                   ["]+THIS.TargetFileName+["]+ ;
                   GS_COMMAND_STRING_2 + ;
                   + lcFileString   
      * forward slashes or doublebackslashes
      lcContents = STRTRAN(lcContents,"\","/")                 
      IF FILE(lcFile)
         ERASE (lcFile)
      ENDIF
      liH = FCREATE(lcFile)
      FPUTS(liH,lcContents)
      *&* second param of FFLUSH removed, not available in early versions
      FFLUSH(liH)      
      FCLOSE(liH)
*      STRTOFILE(lcContents, lcFile) seemed to cause some problems even in VFP 9
   RETURN lcFile
   
   PROTECTED PROCEDURE GetQuotedFileString()
      LOCAL liIndex, lcReturn, lcFile
      lcReturn = ""
      FOR liIndex = 1 TO ALEN(THIS.reportTempFiles)
         IF NOT EMPTY(THIS.reportTempFiles[liIndex])
            lcFile = ALLTRIM(SUBSTR(THIS.reportTempFiles[liIndex],RAT("\",THIS.reportTempFiles[liIndex]) + 1))
            IF LEFT(lcFile,1) # ["]
               lcFile = ["]+lcFile 
            ENDIF   
            IF RIGHT(lcFile,1) # ["]
               lcFile = lcFile + ["]
            ENDIF   
            lcFile = " " + lcFile + " "
            lcReturn = lcReturn + lcFile
         ENDIF
      ENDFOR
   RETURN lcReturn
   
   PROCEDURE LoadPrinterInfo()
      IF NOT THIS.VerifyGSLibrary()
         RETURN .F.
      ENDIF   
      IF NOT THIS.VerifyPrinterSetup()
         RETURN .F.
      ENDIF
      IF NOT THIS.AdjustVFPPrinterSetups()
         RETURN .F.
      ENDIF
   ENDPROC
   
   PROTECTED PROCEDURE UnloadPrinterInfo()
      *&* code removed
   ENDPROC   
   
   PROTECTED PROCEDURE PrepareFRXPrintInfo()
      *&* code removed
   ENDPROC

   PROTECTED PROCEDURE AdjustVFPPrinterSetups()
      LOCAL llFailure, oError AS Exception, liSelect
      liSelect = SELECT()
      THIS.prepareFRXPrintInfo()
*&*      SELECT (THIS.FRXPrintInfoAlias) not available
*&*      SYS(1037,2)
      SET PRINTER TO NAME (THIS.PSDriverSetupName)
      SELECT (liSelect)
   ENDPROC
   
   PROCEDURE VerifyPrinterSetup()
      PRIVATE laSetups[1]
      LOCAL lcCmd, liTrials,oShell, lcFailureConditions
      lcFailureConditions = ;
       " APRINTERS(laSetups) = 0 OR " + ;
       " ASCAN(laSetups,["+THIS.PSDriverSetupName+"],1,ALEN(laSetups)) = 0 "  
      
      * check for the existance of the appropriate printer driver
      * if it is not there, attempt to install
      * if it is not installed after attempt,
      * return .F.      
     IF &lcFailureConditions.
        IF NOT THIS.QuietMode
           WAIT WINDOW NOWAIT INSTALLING_DRIVER_LOC
        ENDIF           
        lcCmd = [%windir%\\system32\\rundll32.exe  printui.dll,PrintUIEntry /if /b ] + ;
                ["] + THIS.PSDriverSetupName + ["] + ;
                [ /f %windir%\\inf\\ntprint.inf /r  "lpt1:" /m "] + ;
                DRIVER_TO_USE + [" /Z]
        llReturn = THIS.ProgExecute(lcCmd,lcFailureConditions,THIS.WaitAttempts)              
        IF NOT THIS.QuietMode
           WAIT CLEAR
        ENDIF
     ENDIF   
     RETURN APRINTERS(laSetups) > 0 && AND ;
           ASCAN(laSetups,THIS.PSDriverSetupName,1,ALEN(laSetups)) > 0   
          *&* TBD: to do this properly, you should upper all the Setups and
          *&* check them all case-insensitive
*!*  VFP 9:       ASCAN(laSetups,THIS.PSDriverSetupName,1,ALEN(laSetups,1),1,7) > 0   
   ENDPROC
   
   PROCEDURE VerifyGSLibrary()
      IF EMPTY(SYS(2000, THIS.GSLocation + THIS.GSExecFile,THIS.GSLocation))
         * creating directory if necessary/possible.
         IF NOT DIRECTORY(THIS.GSLocation)
            MD (THIS.GSLocation)
         ENDIF
         IF DIRECTORY(THIS.GSLocation) AND FILE("Vfp5Install.DBF")
            * attempt to install
            IF NOT THIS.QuietMode
               WAIT WINDOW NOWAIT INSTALLING_LIBRARY_LOC
            ENDIF               
            LOCAL liSelect, lcFile
            liSelect = SELECT()
            SELECT 0
            USE Vfp5Install
            SCAN
               lcFile = THIS.GSLocation + ALLTRIM(dir)+ ALLTRIM(filename)
               IF NOT DIRECTORY(THIS.GSLocation + ALLTRIM(dir))
                  MD (THIS.GSLocation + ALLTRIM(dir))
               ENDIF
               IF NOT FILE(lcFile)
                  COPY MEMO Contents TO (lcFile)
               ENDIF
            ENDSCAN
            USE 
            SELECT(liSelect)
            IF NOT THIS.QuietMode
               WAIT CLEAR
            ENDIF
         ENDIF
      ENDIF   
      RETURN NOT EMPTY(SYS(2000,THIS.GSLocation + THIS.GSExecFile))
   ENDPROC
   
*!*   PROCEDURE GSLocation_Verify(tcVal)
*!*      IF VARTYPE(tcVal) = "C" AND NOT EMPTY(tcVal)
*!*        THIS.GSLocation = UPPER(ADDBS(ALLTRIM(tcVal)))
*!*      ENDIF
*!*   ENDPROC

*!*	   PROCEDURE PSDriverSetupName_Assign(tcVal)
*!*	      IF VARTYPE(tcVal) = "C" AND NOT EMPTY(tcVal)
*!*	         LOCAL laSetups[1], liIndex, llFoundAndNotAppropriate
*!*	         FOR liIndex = 1 TO APRINTERS(laSetups) 
*!*	             IF ( UPPER(laSetups[liIndex,1]) == ;
*!*	                  UPPER(ALLTRIM(THIS.PSDriverSetupName)) ) AND ;
*!*	                (NOT ( UPPER(laSetups[liIndex,3]) == ;
*!*	                       UPPER(ALLTRIM(DRIVER_TO_USE)) ) )
*!*	                 llFoundAndNotAppropriate = .T.
*!*	                 EXIT
*!*	             ENDIF 
*!*	         ENDFOR
*!*	         IF NOT llFoundAndNotAppropriate
*!*	           THIS.PSDriverSetupName = ALLTRIM(tcVal)
*!*	         ENDIF
*!*	      ENDIF
*!*	   ENDPROC

   
    PROTECTED PROCEDURE ProgExecute(zCmdLine, cFailureConditions, nWait, cStatus, lUseRun)
       LOCAL liTrials, lcConditions, lcCommand
       IF NOT EMPTY(cFailureConditions)
          lcConditions = "("+cFailureConditions + ") AND "
       ELSE
          lcConditions = ""   
       ENDIF
       IF NOT (EMPTY(cStatus) OR THIS.QuietMode) &&  OR THIS.IsOle)
          WAIT WINDOW NOWAIT cStatus
       ENDIF
       lcCommand = zCmdLine
       IF AT("%windir%",lcCommand) > 0  && case sensitive, we don't want to upper the whole string
                                        && so you should be careful what version
                                        && of this environmental value you pass
          lcCommand = STRTRAN(lcCommand,"%windir%", GETENV("windir"))                                        
       ENDIF
       lcCommand = ALLTR(lcCommand)
       IF lUseRun
          RUN /N2 &lcCommand.
       ELSE
          LOCAL lcCmd, lcArgs, liPos
          liPos = AT(" ",lcCommand)
          IF  liPos = 0
             lcCmd = lcCommand
             lcArgs = ""
          ELSE
             lcCmd = LEFT(lcCommand,liPos -1)
             lcArgs = SUBSTR(lcCommand, liPos + 1)
          ENDIF
          DECLARE INTEGER ShellExecute ;
                   IN SHELL32.DLL ;
                   INTEGER nWinHandle,;
                   STRING cOperation,;   
                   STRING cFileName,;
                   STRING cParameters,;
                   STRING cDirectory,;
                   INTEGER nShowWindow
          ShellExecute(0,"",lcCmd,lcArgs,"",0)                    
       ENDIF          
       IF TYPE("nWait") = "N"
           DECLARE Sleep IN Win32API INTEGER   
           liTrials = 0
           DO WHILE &lcConditions. liTrials < 100
              SLEEP(1000)
              liTrials = liTrials + 1
           ENDDO   
         ENDIF   
        IF NOT (EMPTY(cStatus) OR THIS.QuietMode) &&  OR THIS.IsOle)
           WAIT CLEAR
        ENDIF
        IF EMPTY(cFailureConditions)
           RETURN .T.
        ELSE      
           RETURN (NOT &cFailureConditions.)
        ENDIF

    ENDPROC
    
    PROTECTED PROCEDURE PrepareErrorMessage
       LPARAMETERS m.nError, m.cMethod, m.nLine, m.cName, m.cMessage, m.cCodeLine
       LOCAL m.lcErrorMessage, m.lcCodeLineMsg

       IF TYPE("cMessage") = "C"
          m.lcErrorMessage = m.cMessage
       ELSE
          m.lcErrorMessage = MESSAGE()
       ENDIF
       m.lcErrorMessage = m.lcErrorMessage + CHR(13) + CHR(13)

       IF TYPE("cName") = "C"
          m.lcErrorMessage = m.lcErrorMessage + m.cName
       ELSE
          m.lcErrorMessage = m.lcErrorMessage + this.Name
       ENDIF

       m.lcErrorMessage = m.lcErrorMessage + CHR(13)+ ;
     		 OUTPUTCLASS_ERRNOLABEL_LOC +ALLTRIM(STR(m.nError))+CHR(13)+ ;
       		 OUTPUTCLASS_ERRPROCLABEL_LOC +LOWER(ALLTRIM(m.cMethod))
		
       IF TYPE("m.cCodeLine") = "C"
          m.lcCodeLineMsg = m.cCodeLine
       ELSE
          m.lcCodeLineMsg = MESSAGE(1)
       ENDIF		

       IF BETWEEN(m.nLine,1,100000) AND NOT m.lcCodeLineMsg="..."
       	m.lcErrorMessage= ;
              m.lcErrorMessage+CHR(13)+ OUTPUTCLASS_ERRLINELABEL_LOC+ ;
       	    ALLTRIM(STR(m.nLine))
       	IF NOT EMPTY(m.lcCodeLineMsg)
              m.lcErrorMessage= ;
       		   m.lcErrorMessage+CHR(13)+CHR(13)+m.lcCodeLineMsg
       	ENDIF
       ENDIF

       RETURN m.lcErrorMessage    
    ENDPROC
    
    PROTECTED PROC VerifyTargetFile
       LOCAL m.lcFile

       m.lcFile =  ALLTR(CHRTRAN(THIS.TargetFileName,;
                       OUTPUTCLASS_FILENAME_CHARS_DISALLOWED,"_"))

       IF NOT DIRECTORY(SUBSTR(m.lcFile,1,RAT("\",m.lcFile)))
          m.lcFile = FULLPATH(ALLTR(m.lcFile))
       ENDIF   

       IF DIRECTORY(m.lcFile)
          * we have to generate a filename
          IF RIGHT(m.lcFile,1) # "\"
             m.lcFile = m.lcFile + "\"
          ENDIF
          m.lcFile = m.lcFile + SYS(2015)
       ENDIF

       THIS.TargetFileName = m.lcFile   

       IF AT(".",THIS.TargetFileName) = 0 AND ;
          RIGHT(THIS.TargetFileName,1) # "."      
       
          THIS.TargetFileExt = CHRTRAN(THIS.TargetFileExt,;
                        OUTPUTCLASS_FILENAME_CHARS_DISALLOWED,"_")

          THIS.TargetFileName = THIS.TargetFileName + THIS.TargetFileExt
       ENDIF                                 
 
       IF FILE(THIS.TargetFileName)
          ERASE (THIS.TargetFileName) NORECYCLE
       ENDIF
   ENDPROC       
    
    
ENDDEFINE

